@@ -3,11 +3,11 @@
 /* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */
 
 /**
- * PHPUnit3.2 test framework script for the Crypt_GPG package.
+ * PHPUnit AllTests suite for the Services_Akismet2 package.
  *
- * These tests require the PHPUnit 3.2 package to be installed. PHPUnit is
- * installable using PEAR. See the
- * {@link http://www.phpunit.de/pocket_guide/3.2/en/installation.html manual}
+ * These tests require the PHPUnit 3.6 or greater package to be installed.
+ * PHPUnit is installable using PEAR. See the
+ * {@link http://www.phpunit.de/manual/3.6/en/installation.html manual}
  * for detailed installation instructions.
  *
  * LICENSE:
@@ -29,9 +29,9 @@
  * @category  Encryption
  * @package   Crypt_GPG
  * @author    Michael Gauthier <mike@silverorange.com>
- * @copyright 2005-2011 silverorange
+ * @copyright 2005-2013 silverorange
  * @license   http://www.gnu.org/copyleft/lesser.html LGPL License 2.1
- * @version   CVS: $Id: TestCase.php 316438 2011-09-09 03:58:13Z gauthierm $
+ * @version   CVS: $Id$
  * @link      http://pear.php.net/package/Crypt_GPG
  */
 
@@ -91,11 +91,6 @@ abstract class Crypt_GPG_TestCase extends PHPUnit_Framework_TestCase
 
     protected $gpg;
 
-    // }}}
-    // {{{ private properties
-
-    private $_old_error_level;
-
     // }}}
     // {{{ getOptions()
 
@@ -103,7 +98,7 @@ abstract class Crypt_GPG_TestCase extends PHPUnit_Framework_TestCase
     {
         return array(
             'homedir' => dirname(__FILE__) . '/' . self::HOMEDIR,
-            'debug'   => true
+//            'debug'   => true
         );
     }
 
@@ -114,8 +109,6 @@ abstract class Crypt_GPG_TestCase extends PHPUnit_Framework_TestCase
 
     public function setUp()
     {
-        $this->_old_error_level = error_reporting(E_ALL | E_STRICT);
-
         // load test configuration file if it exists
         $configFilename = dirname(__FILE__).'/config.php';
         if (file_exists($configFilename)) {
@@ -571,8 +564,6 @@ TEXT;
 
         $this->_tearDownKeyring();
         $this->_tearDownTempdir();
-
-        error_reporting($this->_old_error_level);
     }
 
     // }}}
@@ -596,6 +587,16 @@ TEXT;
             }
         }
 
+        $dirnames = array(
+            $this->getKeyringFilename('private-keys-v1.d')
+        );
+
+        foreach ($dirnames as $dirname) {
+            if (file_exists($dirname)) {
+                rmdir($dirname);
+            }
+        }
+
         rmdir(dirname(__FILE__) . '/' . self::HOMEDIR);
     }
 
@@ -611,7 +612,11 @@ TEXT;
         foreach ($iterator as $file) {
             if (!$file->isDot()) {
                 $filename = $this->getTempFilename($file->getFilename());
-                unlink($filename);
+                if (is_dir($filename)) {
+                    rmdir($filename);
+                } else {
+                    unlink($filename);
+                }
             }
         }
 
